body {
  /*background-color: #4A849F;
  color: #FFFFFF;*/
}

h1 {
  text-align: center;
}
h3 {
  text-align: center;
}

.alignment-header {
  text-align: center;
}
.carousel.alignment-carousel {
  margin: 30px;
}
.alignment-carousel .d-block {
  height: 450px;
  width: auto !important;
  margin: 10px auto;
}
.alignment-carousel ol.carousel-indicators {
  margin-bottom: 0px;
  bottom: -10px;
}

.figure-row {
  margin: 20px;
  display: flex;
  flex-direction: row;

}
.figure-col {
  width: 50%;
}
img.figure-img {
  max-width: 100%;
  padding: 10px;
}